package avrasm;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;

public class AvrParser {

	String filename;
	List<Label> labels = new LinkedList<Label>();
	List<Instruction> instructions = new LinkedList<Instruction>();
	List<Define> defines = new LinkedList<Define>();

	private Integer getRegisterIndex(String operand, Integer lineNumber) throws AVRException {

		if (operand.compareTo("x") == 0)
			return 12;   // code taken from ISA datasheet for ld
		else if (operand.compareTo("x+") == 0)
			return 13;   // code taken from ISA datasheet for ld
		else if (operand.compareTo("-x") == 0)
			return 14;   // code taken from ISA datasheet for ld
		else if (operand.compareTo("y") == 0)
			return 8;    // code taken from ISA datasheet for ld
		else if (operand.compareTo("y+") == 0)
			return 9;    // code taken from ISA datasheet for ld
		else if (operand.compareTo("-y") == 0)
			return 10;   // code taken from ISA datasheet for ld
		else if (operand.compareTo("z") == 0)
			return 0;    // code taken from ISA datasheet for ld
		else if (operand.compareTo("z+") == 0)
			return 1;    // code taken from ISA datasheet for ld
		else if (operand.compareTo("-z") == 0)
			return 2;    // code taken from ISA datasheet for ld
		else if (operand.compareTo("zh") == 0)
			return 31;
		else if (operand.compareTo("zl") == 0)
			return 30;
		else if (operand.compareTo("yh") == 0)
			return 29;
		else if (operand.compareTo("yl") == 0)
			return 28;
		else if (operand.compareTo("xh") == 0)
			return 27;
		else if (operand.compareTo("xl") == 0)
			return 26;
		else if (operand.charAt(0) == 'r')
			return Integer.parseInt(operand.substring(1));
		else
			throw new AVRException(ErrorType.INVALID_OPERAND, operand, lineNumber);
	}

	private String getRegister(String operand, Integer lineNumber) throws AVRException {
		return Integer.toBinaryString(getRegisterIndex(operand, lineNumber));
	}

	private String getBinaryForm(int operand, int lineNumber) throws AVRException {
		try {
			return Integer.toBinaryString(operand);
		} catch (NumberFormatException ex) {
			throw new AVRException(ErrorType.INVALID_IMMEDIATE, Integer.toString(operand), lineNumber);
		}
	}

	private String parseConstant(String string) {

		// if it's an ascii character
		if (string.matches("'(.+)'")) {
			byte[] bytes = string.getBytes();
			return Integer.toString((int) bytes[1]);
		// if it's a hexadecimal value
		} else if (string.startsWith("0x")) {
			int value = Integer.parseInt(string.substring(2), 16);
			return Integer.toString(value);
		} else if (string.startsWith("0b")) {
			int value = Integer.parseInt(string.substring(2), 2);
			return Integer.toString(value);
		} else {
			return string;
		}
	}

	private String replaceDefine(String string) {

		for (Define define : defines) {
			if (define.getLvalue().compareTo(string) == 0) {
				return define.getRvalue();
			}
		}
		return string;
	}

	private String signExtend(String value, int numBits) {

		if (numBits < value.length()) {
			return value.substring(value.length() - numBits);
		}
		char bit = value.charAt(0);
		char[] chars = new char[numBits - value.length()];
		Arrays.fill(chars, bit);
		String padding = new String(chars);
		return padding + value;
	}

	private String extend(String value, int numBits) {
		if (value.length() > numBits) {
			return value.substring(value.length() - numBits);
		} else {
			return String.format("%" + numBits + "s", value).replace(" ", "0");
		}
	}

	private String convertRegImmediateForm(String opcode, List<String> operands,
			int immediateBits, int lineNumber) throws AVRException {

		if (operands.size() != 2)
			throw new AVRException(ErrorType.INVALID_OPERAND_NUMBER, opcode, lineNumber);

		String immediate;
		String imdPart1;
		String imdPart2;
		String rd;
		int rdInt;

		immediate = operands.get(1);
		immediate = getUsableImmediate(immediate, immediateBits, lineNumber);
		// split it in two
		int middle = immediate.length() - 4;
		imdPart1 = immediate.substring(0, middle);
		imdPart2 = immediate.substring(middle);
		rd = getRegister(operands.get(0), lineNumber);
		rdInt = Integer.parseInt(rd, 2);
		if (rdInt < 16) {
			throw new AVRException(ErrorType.INVALID_OPERAND, operands.get(0), lineNumber);
		}
		rd = Integer.toBinaryString(rdInt - 16);
		rd = extend(rd, 4);
		return opcode + imdPart1 + rd + imdPart2;
	}

	private String getUsableImmediate(String operand, int numBits, int lineNumber) throws AVRException {
		String immediate;
		int imd;

		// it can be "'a'", for example
		immediate = replaceDefine(operand);
		// transform the string is in decimal form
		immediate = parseConstant(immediate);
		// get the value into an integer
		try {
			imd = Integer.parseInt(immediate);
		} catch (NumberFormatException ex) {
			throw new AVRException(ErrorType.INVALID_IMMEDIATE, operand, lineNumber);
		}
		// transform the integer back in string, but in binary form
		immediate = getBinaryForm(imd, lineNumber);
		// make sure the sign doesn't get messed up
		if (imd >= 0)
			immediate = extend(immediate, numBits);
		else
			immediate = signExtend(immediate, numBits);
		return immediate;
	}

	private String convertTwoRegForm(String opcode, List<String> operands, int lineNumber) throws AVRException {

		if (operands.size() != 2)
			throw new AVRException(ErrorType.INVALID_OPERAND_NUMBER, opcode, lineNumber);

		String rr, rd;
		String rrPart1, rrPart2;

		rd = extend(getRegister(operands.get(0), lineNumber), 5);
		rr = extend(getRegister(operands.get(1), lineNumber), 5);
		rrPart1 = rr.substring(0, 1);
		rrPart2 = rr.substring(1, 5);
		return opcode + rrPart1 + rd + rrPart2;
	}

	private String convertLDI(List<String> operands, int lineNumber) throws AVRException {
		// 1110 KKKK dddd KKKK
		System.out.println("converting ldi at line " + lineNumber);
		return convertRegImmediateForm("1110", operands, 8, lineNumber);
	}

	private String convertCP(List<String> operands, int lineNumber) throws AVRException {
		// 0001 01rd dddd rrrr
		System.out.println("converting cp at line " + lineNumber);
		return convertTwoRegForm("000101", operands, lineNumber);
	}

	private String convertCPI(List<String> operands, int lineNumber) throws AVRException {
		// 0011 KKKK dddd KKKK
		System.out.println("converting cpi at line " + lineNumber);
		return convertRegImmediateForm("0011", operands, 8, lineNumber);
	}

	private String convertLD(List<String> operands, int lineNumber) throws AVRException {

		// ld Rd, x   : 1001 000d dddd 1100
		// ld Rd, x+  : 1001 000d dddd 1101
		// ld Rd, -x  : 1001 000d dddd 1110
		// ldd Rd, y  : 1000 000d dddd 1000
		// ldd Rd, y+ : 1001 000d dddd 1001
		// ldd Rd, -y : 1001 000d dddd 1010
		System.out.println("converting ld at line " + lineNumber);

		if (operands.size() != 2)
			throw new AVRException(ErrorType.INVALID_OPERAND_NUMBER, "ld", lineNumber);

		String opcode1 = "1001000";
		String opcode2 = "1000000";
		String opcode;
		String rd = getRegister(operands.get(0), lineNumber);
		String indirect = getRegister(operands.get(1), lineNumber);
		if (indirect.compareTo("1000") == 0) {        // y
			opcode = opcode2;
		} else if (indirect.compareTo("1100") == 0 || // x
		           indirect.compareTo("1101") == 0 || // x+
		           indirect.compareTo("1110") == 0 || // -x
		           indirect.compareTo("1001") == 0 || // y+
		           indirect.compareTo("1010") == 0) { // -y
			opcode = opcode1;
		} else {
			throw new AVRException(ErrorType.INVALID_OPERAND, operands.get(0), lineNumber);
		}
		return opcode + rd + indirect;
	}

	private String convertLDD(List<String> operands, int lineNumber) throws AVRException {
		throw new AVRException(ErrorType.UNIMPLEMENTED, "ldd", lineNumber);
	}

	private String convertLDS(List<String> operands, int lineNumber) throws AVRException {
		// 1010 0kkk dddd kkkk
		System.out.println("converting lds at line " + lineNumber);
		return convertRegImmediateForm("10100", operands, 7, lineNumber);
	}

	private String convertST(List<String> operands, int lineNumber) throws AVRException {

		// st x, Rr  : 1001 001r rrrr 1100
		// st x+, Rr : 1001 001r rrrr 1101
		// st -x, Rr : 1001 001r rrrr 1110
		// st y, Rr  : 1000 001r rrrr 0000
		// st y+, Rr : 1001 001r rrrr 1001
		// st -y, Rr : 1001 001r rrrr 1010
		System.out.println("converting st at line " + lineNumber);

		if (operands.size() != 2)
			throw new AVRException(ErrorType.INVALID_OPERAND_NUMBER, "st", lineNumber);

		String opcode1 = "1001001";
		String opcode2 = "1000001";
		String opcode;

		String rr = getRegister(operands.get(1), lineNumber);
		String indirect = getRegister(operands.get(0), lineNumber);
		if (indirect.compareTo("0000") == 0) {        // y
			opcode = opcode2;
		} else if (indirect.compareTo("1100") == 0 || // x
		           indirect.compareTo("1101") == 0 || // x+
		           indirect.compareTo("1110") == 0 || // -x
		           indirect.compareTo("1001") == 0 || // y+
		           indirect.compareTo("1010") == 0) { // -y
			opcode = opcode1;
		} else {
			throw new AVRException(ErrorType.INVALID_OPERAND, operands.get(0), lineNumber);
		}
		return opcode + rr + indirect;
	}

	private String convertSTD(List<String> operands, int lineNumber) throws AVRException {

		System.out.println("converting std at line " + lineNumber);
		throw new AVRException(ErrorType.UNIMPLEMENTED, "std", lineNumber);
	}

	private String convertSTS(List<String> operands, int lineNumber) throws AVRException {
		// 1010 1kkk dddd kkkk
		System.out.println("converting sds at line " + lineNumber);

		/* Instruction appears in code as "sts addr, register"
		 * invert operand order so it conforms to reg-immediate form.
		 */
		if (operands.size() != 2)
			throw new AVRException(ErrorType.INVALID_OPERAND_NUMBER, "sts", lineNumber);
		String temp = operands.get(0);
		operands.set(0, operands.get(1));
		operands.set(1, temp);

		return convertRegImmediateForm("10101", operands, 7, lineNumber);
	}

	private String convertMOV(List<String> operands, int lineNumber) throws AVRException {
		// 0010 11rd dddd rrrr
		System.out.println("converting mov at line " + lineNumber);
		return convertTwoRegForm("001011", operands, lineNumber);
	}


	private String convertRJMP(List<String> operands, int instrNumber, int lineNumber) throws AVRException {

		// 1100 kkkk kkkk kkkk
		System.out.println("converting rjmp at line " + lineNumber);

		if (operands.size() != 1)
			throw new AVRException(ErrorType.INVALID_OPERAND_NUMBER, "rjmp", lineNumber);

		String immediate = null;
		String destination = operands.get(0);
		String opcode = "1100";
		int imd = -instrNumber - 1;

		for (Label label : labels) {
			if (destination.compareTo(label.getName()) == 0) {
				imd = imd + label.getInstructionNumber();
				immediate = Integer.toString(imd);
				immediate = getUsableImmediate(immediate, 12, lineNumber);
			}
		}
		if (immediate == null) {
			throw new AVRException(ErrorType.UNDEFINED_LABEL, operands.get(0), lineNumber);
		}
		return opcode + immediate;
	}

	private String convertRCALL(List<String> operands, int instrNumber, int lineNumber) throws AVRException {
		// 1101 kkkk kkkk kkkk
		System.out.println("converting rcall at line " + lineNumber);

		if (operands.size() != 1) {
			throw new AVRException(ErrorType.INVALID_OPERAND_NUMBER, "rcall", lineNumber);
		}

		String immediate = null;
		String destination = operands.get(0);
		String opcode = "1101";
		int imd = -instrNumber - 1;

		for (Label label : labels) {
			if (destination.compareTo(label.getName()) == 0) {
				imd = imd + label.getInstructionNumber();
				immediate = Integer.toString(imd);
				immediate = getUsableImmediate(immediate, 12, lineNumber);
			}
		}
		if (immediate == null) {
			throw new AVRException(ErrorType.UNDEFINED_LABEL, operands.get(0), lineNumber);
		}
		return opcode + immediate;
	}

	private String convertBranch(List<String> operands, boolean condition,
		int sreg_bit, int instrNumber, int lineNumber) throws AVRException {

		if (operands.size() != 1) {
			throw new AVRException(ErrorType.INVALID_OPERAND_NUMBER, "branch", lineNumber);
		}

		// brbc: 1111 01kk kkkk ksss
		// brbs: 1111 00kk kkkk ksss
		String opcode = (condition == true) ? "111100" : "111101";
		String bit = String.format("%3s", Integer.toBinaryString(sreg_bit)).replace(" ", "0");
		String destination = operands.get(0);
		String immediate = null;
		int imd = -instrNumber - 1;

		for (Label label : labels) {
			if (destination.compareTo(label.getName()) == 0) {
				imd = imd + label.getInstructionNumber();
				immediate = Integer.toString(imd);
				immediate = getUsableImmediate(immediate, 7, lineNumber);
			}
		}
		if (immediate == null) {
			throw new AVRException(ErrorType.UNDEFINED_LABEL, operands.get(0), lineNumber);
		}
		return opcode + immediate + bit;
	}

	private String convertBRCC(List<String> operands, int instrNumber, int lineNumber) throws AVRException {
		System.out.println("converting brcc at line " + lineNumber);
		// C flag is 0
		return convertBranch(operands, false, 0, instrNumber, lineNumber);
	}

	private String convertBRCS(List<String> operands, int instrNumber, int lineNumber) throws AVRException {
		System.out.println("converting brcs at line " + lineNumber);
		// C flag is 0
		return convertBranch(operands, true, 0, instrNumber, lineNumber);
	}

	private String convertBREQ(List<String> operands, int instrNumber, int lineNumber) throws AVRException {
		System.out.println("converting breq at line " + lineNumber);
		// Z flag is 1
		return convertBranch(operands, true, 1, instrNumber, lineNumber);
	}

	private String convertBRNE(List<String> operands, int instrNumber, int lineNumber) throws AVRException {
		System.out.println("converting brne at line " + lineNumber);
		// Z flag is 0
		return convertBranch(operands, false, 1, instrNumber, lineNumber);
	}

	private String convertBRMI(List<String> operands, int instrNumber, int lineNumber) throws AVRException {
		System.out.println("converting brmi at line " + lineNumber);
		// N flag is 2
		return convertBranch(operands, true, 2, instrNumber, lineNumber);
	}

	private String convertBRPL(List<String> operands, int instrNumber, int lineNumber) throws AVRException {
		System.out.println("converting brpl at line " + lineNumber);
		// N flag is 2
		return convertBranch(operands, false, 2, instrNumber, lineNumber);
	}

	private String convertBRVC(List<String> operands, int instrNumber, int lineNumber) throws AVRException {
		System.out.println("converting brvc at line " + lineNumber);
		// V flag is 3
		return convertBranch(operands, false, 3, instrNumber, lineNumber);
	}

	private String convertBRVS(List<String> operands, int instrNumber, int lineNumber) throws AVRException {
		System.out.println("converting brvs at line " + lineNumber);
		// V flag is 3
		return convertBranch(operands, true, 3, instrNumber, lineNumber);
	}

	private String convertOneRegType(String opcodePart1, String opcodePart2,
		List<String> operands, int lineNumber) throws AVRException {

		if (operands.size() != 1) {
			throw new AVRException(ErrorType.INVALID_OPERAND_NUMBER, "op", lineNumber);
		}

		String rd = getRegister(operands.get(0), lineNumber);
		rd = extend(rd, 5);
		return opcodePart1 + rd + opcodePart2;
	}

	private String convertNEG(List<String> operands, int lineNumber) throws AVRException {
		// 1001 010d dddd 0001
		System.out.println("converting neg at line " + lineNumber);
		return convertOneRegType("1001010", "0001", operands, lineNumber);
	}

	private String convertPUSH(List<String> operands, int lineNumber) throws AVRException {
		// 1001 001d dddd1111
		System.out.println("converting push at line " + lineNumber);
		return convertOneRegType("1001001", "1111", operands, lineNumber);
	}

	private String convertPOP(List<String> operands, int lineNumber) throws AVRException {
		// 1001 000d dddd 1111
		System.out.println("converting pop at line " + lineNumber);
		return convertOneRegType("1001000", "1111", operands, lineNumber);
	}

	private String convertAND(List<String> operands, int lineNumber) throws AVRException {
		// 0010 00rd dddd rrrr
		System.out.println("converting and at line " + lineNumber);
		return convertTwoRegForm("001000", operands, lineNumber);
	}

	private String convertANDI(List<String> operands, int lineNumber) throws AVRException {
		// 0111 KKKK dddd KKKK
		System.out.println("converting andi at line " + lineNumber);
		return convertRegImmediateForm("0111", operands, 8, lineNumber);
	}

	private String convertOR(List<String> operands, int lineNumber) throws AVRException {
		// 0010 10rd dddd rrrr
		System.out.println("converting or at line " + lineNumber);
		return convertTwoRegForm("001010", operands, lineNumber);
	}

	private String convertORI(List<String> operands, int lineNumber) throws AVRException {
		// 0110 KKKK dddd KKKK
		System.out.println("converting ori at line " + lineNumber);
		return convertRegImmediateForm("0110", operands, 8, lineNumber);
	}

	private String convertEOR(List<String> operands, int lineNumber) throws AVRException {
		// 0010 01rd dddd rrrr
		System.out.println("converting eor at line " + lineNumber);
		return convertTwoRegForm("001001", operands, lineNumber);
	}

	private String convertSBICBI(String opcode, List<String> operands, int lineNumber) throws AVRException {
		if (operands.size() != 2)
			throw new AVRException(ErrorType.INVALID_OPERAND_NUMBER, opcode, lineNumber);

		String bits;
		String immediate;

		immediate = operands.get(0);
		immediate = getUsableImmediate(immediate, 5, lineNumber);
		bits      = operands.get(1);
		bits      = getUsableImmediate(bits, 3, lineNumber);
		return opcode + immediate + bits;
	}

	private String convertSBI(List<String> operands, int lineNumber) throws AVRException {
		// 1001 1010 AAAA Abbb
		System.out.println("converting sbi at line " + lineNumber);
		return convertSBICBI("10011010", operands, lineNumber);
	}

	private String convertCBI(List<String> operands, int lineNumber) throws AVRException {
		// 1001 1000 AAAA Abbb
		System.out.println("converting cbi at line " + lineNumber);
		return convertSBICBI("10011000", operands, lineNumber);
	}

	private String convertIN(List<String> operands, int lineNumber) throws AVRException {
		// 1011 0AAd dddd AAAA
		System.out.println("converting in at line " + lineNumber);

		if (operands.size() != 2) {
			throw new AVRException(ErrorType.INVALID_OPERAND_NUMBER, "in", lineNumber);
		}

		String opcode = "10110";
		String immediate;
		String immediatePart1, immediatePart2;
		String rd;

		immediate = operands.get(1);
		immediate = getUsableImmediate(immediate, 6, lineNumber);
		immediatePart1 = immediate.substring(0, 2);
		immediatePart2 = immediate.substring(2, 6);

		rd = extend(getRegister(operands.get(0), lineNumber), 5);
		return opcode + immediatePart1 + rd + immediatePart2;
	}

	private String convertOUT(List<String> operands, int lineNumber) throws AVRException {
		// 1011 1AAd dddd AAAA
		System.out.println("converting out at line " + lineNumber);

		if (operands.size() != 2) {
			throw new AVRException(ErrorType.INVALID_OPERAND_NUMBER, "in", lineNumber);
		}

		String opcode = "10111";
		String immediate;
		String immediatePart1, immediatePart2;
		String rd;

		immediate = operands.get(0);
		immediate = getUsableImmediate(immediate, 6, lineNumber);
		immediatePart1 = immediate.substring(0, 2);
		immediatePart2 = immediate.substring(2, 6);

		rd = extend(getRegister(operands.get(1), lineNumber), 5);
		return opcode + immediatePart1 + rd + immediatePart2;
	}

	private String convertADD(List<String> operands, int lineNumber) throws AVRException {
		// 0000 11rd dddd rrrr
		System.out.println("converting add at line " + lineNumber);
		return convertTwoRegForm("000011", operands, lineNumber);
	}

	private String convertADC(List<String> operands, int lineNumber) throws AVRException {
		// 0001 11rd dddd rrrr
		System.out.println("converting adc at line " + lineNumber);
		return convertTwoRegForm("000111", operands, lineNumber);
	}

	private String convertSUB(List<String> operands, int lineNumber) throws AVRException {
		// 0001 10rd dddd rrrr
		System.out.println("converting sub at line " + lineNumber);
		return convertTwoRegForm("000110", operands, lineNumber);
	}

	private String convertSUBI(List<String> operands, int lineNumber) throws AVRException {
		// 0101 KKKK dddd KKKK
		System.out.println("converting subi at line " + lineNumber);
		return convertRegImmediateForm("0101", operands, 8, lineNumber);
	}

	private String convertSBC(List<String> operands, int lineNumber) throws AVRException {
		// 0000 10rd dddd rrrr
		System.out.println("converting sbc at line " + lineNumber);
		return convertTwoRegForm("000010", operands, lineNumber);
	}

	private String convertSBCI(List<String> operands, int lineNumber) throws AVRException {
		// 0100 KKKK dddd KKKK
		System.out.println("converting sbci at line " + lineNumber);
		return convertRegImmediateForm("0100", operands, 8, lineNumber);
	}

	private String convertINC(List<String> operands, int lineNumber) throws AVRException {
		// 1001 010d dddd 0011
		System.out.println("converting inc at line " + lineNumber);
		return convertOneRegType("1001010", "0011", operands, lineNumber);
	}

	private String convertDEC(List<String> operands, int lineNumber) throws AVRException {
		// 1001 010d dddd 1010
		System.out.println("converting dec at line " + lineNumber);
		return convertOneRegType("1001010", "1010", operands, lineNumber);
	}

	private void convertToBinary(Instruction instr) throws AVRException {
		String mnemonic = instr.getOpcode();
		String binary = null;
		List<String> operands = instr.getOperands();

		if (mnemonic.compareTo("ld") == 0) {
			binary = convertLD(operands, instr.getLineNumber());
		} else if (mnemonic.compareTo("ldd") == 0) {
			binary = convertLDD(operands, instr.getLineNumber());
		} else if (mnemonic.compareTo("lds") == 0) {
			binary = convertLDS(operands, instr.getLineNumber());
		} else if (mnemonic.compareTo("st") == 0) {
			binary = convertST(operands, instr.getLineNumber());
		} else if (mnemonic.compareTo("std") == 0) {
			binary = convertSTD(operands, instr.getLineNumber());
		} else if (mnemonic.compareTo("sts") == 0) {
			binary = convertSTS(operands, instr.getLineNumber());
		} else if (mnemonic.compareTo("ldi") == 0) {
			binary = convertLDI(operands, instr.getLineNumber());
		} else if (mnemonic.compareTo("mov") == 0) {
			binary = convertMOV(operands, instr.getLineNumber());
		} else if (mnemonic.compareTo("brcc") == 0) {
			binary = convertBRCC(operands, instr.getInstructionNumber(), instr.getLineNumber());
		} else if (mnemonic.compareTo("brcs") == 0) {
			binary = convertBRCS(operands, instr.getInstructionNumber(), instr.getLineNumber());
		} else if (mnemonic.compareTo("breq") == 0) {
			binary = convertBREQ(operands, instr.getInstructionNumber(), instr.getLineNumber());
		} else if (mnemonic.compareTo("brne") == 0) {
			binary = convertBRNE(operands, instr.getInstructionNumber(), instr.getLineNumber());
		} else if (mnemonic.compareTo("brvc") == 0) {
			binary = convertBRVC(operands, instr.getInstructionNumber(), instr.getLineNumber());
		} else if (mnemonic.compareTo("brvs") == 0) {
			binary = convertBRVS(operands, instr.getInstructionNumber(), instr.getLineNumber());
		} else if (mnemonic.compareTo("brmi") == 0) {
			binary = convertBRMI(operands, instr.getInstructionNumber(), instr.getLineNumber());
		} else if (mnemonic.compareTo("brpl") == 0) {
			binary = convertBRPL(operands, instr.getInstructionNumber(), instr.getLineNumber());
		} else if (mnemonic.compareTo("rjmp") == 0) {
			binary = convertRJMP(operands, instr.getInstructionNumber(), instr.getLineNumber());
		} else if (mnemonic.compareTo("rcall") == 0) {
			binary = convertRCALL(operands, instr.getInstructionNumber(), instr.getLineNumber());
		} else if (mnemonic.compareTo("ret") == 0) {
			binary = "1001010100001000";
		} else if (mnemonic.compareTo("reti") == 0) {
			binary = "1001010100011000";
		} else if (mnemonic.compareTo("nop") == 0) {
			binary = "0000000000000000";
		} else if (mnemonic.compareTo("sei") == 0) {
			binary = "1001010001111000";
		} else if (mnemonic.compareTo("cli") == 0) {
			binary = "1001010011111000";
		} else if (mnemonic.compareTo("pop") == 0) {
			binary = convertPOP(operands, instr.getLineNumber());
		} else if (mnemonic.compareTo("push") == 0) {
			binary = convertPUSH(operands, instr.getLineNumber());
		} else if (mnemonic.compareTo("in") == 0) {
			binary = convertIN(operands, instr.getLineNumber());
		} else if (mnemonic.compareTo("out") == 0) {
			binary = convertOUT(operands, instr.getLineNumber());
		} else if (mnemonic.compareTo("sbi") == 0) {
			binary = convertSBI(operands, instr.getLineNumber());
		} else if (mnemonic.compareTo("cbi") == 0) {
			binary = convertCBI(operands, instr.getLineNumber());
		} else if (mnemonic.compareTo("inc") == 0) {
			binary = convertINC(operands, instr.getLineNumber());
		} else if (mnemonic.compareTo("dec") == 0) {
			binary = convertDEC(operands, instr.getLineNumber());
		} else if (mnemonic.compareTo("add") == 0) {
			binary = convertADD(operands, instr.getLineNumber());
		} else if (mnemonic.compareTo("adc") == 0) {
			binary = convertADC(operands, instr.getLineNumber());
		} else if (mnemonic.compareTo("sub") == 0) {
			binary = convertSUB(operands, instr.getLineNumber());
		} else if (mnemonic.compareTo("subi") == 0) {
			binary = convertSUBI(operands, instr.getLineNumber());
		} else if (mnemonic.compareTo("sbc") == 0) {
			binary = convertSBC(operands, instr.getLineNumber());
		} else if (mnemonic.compareTo("sbci") == 0) {
			binary = convertSBCI(operands, instr.getLineNumber());
		} else if (mnemonic.compareTo("cp") == 0)  {
			binary = convertCP(operands, instr.getLineNumber());
		} else if (mnemonic.compareTo("cpi") == 0) {
			binary = convertCPI(operands, instr.getLineNumber());
		} else if (mnemonic.compareTo("neg") == 0) {
			binary = convertNEG(operands, instr.getLineNumber());
		} else if (mnemonic.compareTo("and") == 0) {
			binary = convertAND(operands, instr.getLineNumber());
		} else if (mnemonic.compareTo("andi") == 0) {
			binary = convertANDI(operands, instr.getLineNumber());
		} else if (mnemonic.compareTo("or") == 0) {
			binary = convertOR(operands, instr.getLineNumber());
		} else if (mnemonic.compareTo("ori") == 0) {
			binary = convertORI(operands, instr.getLineNumber());
		} else if (mnemonic.compareTo("eor") == 0) {
			binary = convertEOR(operands, instr.getLineNumber());
		}
		if (binary == null)
			throw new AVRException(ErrorType.UNIMPLEMENTED, instr.getOpcode(), instr.getLineNumber());
		instr.setBinaryCode(binary);
	}

	public void generateCode() throws AVRException {
		for (Instruction instr : instructions) {
			convertToBinary(instr);
		}
	}

	private void parseLabel(String label, Integer lineNumber, Integer instructionNumber) throws AVRException {

		for (Label l : labels) {
			if (l.getName().compareTo(label) == 0) {
				throw new AVRException(ErrorType.DUPLICATE_LABEL, label, lineNumber);
			}
		}
		labels.add(new Label(label, lineNumber, instructionNumber));
	}

	private Integer parseInstruction(String instr, Integer lineNumber, Integer instructionNumber) {
		String[] tokens = instr.split("equ");
		if (tokens.length == 2) {
			// instruction is equ
			defines.add(new Define(tokens[0].trim(), tokens[1].trim(), lineNumber));
		} else if (tokens.length == 1) {
			tokens = tokens[0].trim().split("[, \t]");
			List<String> operands = new LinkedList<String>();
			for (int i = 1; i < tokens.length; i++) {
				String operand = tokens[i].trim();
				if (operand.length() != 0)
					operands.add(operand);
			}
			if (tokens[0].length() == 0 && operands.isEmpty())
				return instructionNumber;
			instructions.add(new Instruction(tokens[0].trim(), operands, lineNumber, instructionNumber));
			return instructionNumber + 1;
		}
		return instructionNumber;
	}

	private int parseLine(String line, Integer lineNumber, Integer instructionNumber) throws AVRException {
		/* General assembly line format:
		 * [<label>:] <mnemonic> [<operands>] [;<comment>]
		 */
		line = line.toLowerCase();
		line = line.trim();
		// Step 1. check for comment
		int commentPos = line.indexOf(';');
		if (commentPos != -1) {
			// has comments
			line = line.substring(0, commentPos).trim();
		}
		// Step 2. check for label
		int labelPos = line.indexOf(':');
		if (labelPos != -1) {
			// has label
			String label = line.substring(0, labelPos).trim();
			line = line.substring(labelPos + 1).trim();
			parseLabel(label, lineNumber, instructionNumber);
		}
		if (line.length() != 0) {
			instructionNumber = parseInstruction(line, lineNumber, instructionNumber);
		}
		return instructionNumber;
	}

	public void parse() throws FileNotFoundException, IOException, AVRException {
		BufferedReader br = new BufferedReader(new FileReader(filename));
		String line;
		Integer lineNumber = 1;
		Integer instructionNumber = 0;
		while ((line = br.readLine()) != null) {
			// process the line.
			instructionNumber = parseLine(line, lineNumber, instructionNumber);
			lineNumber++;
		}
		br.close();
	}

	public void write(String outputFileName) throws FileNotFoundException, UnsupportedEncodingException {

		PrintWriter writer = new PrintWriter(outputFileName, "UTF-8");
		for (Instruction instr : instructions) {
			writer.print("\t\t/*\t " + instr.getOpcode() + " \t");
			boolean first = true;
			for (String operand : instr.getOperands()) {
				if (first) {
					writer.print(operand);
					first = false;
				} else {
					writer.print(", " + operand);
				}
			}
			writer.println(" \t\t*/");
			writer.println("\t\t" + instr.getInstructionNumber() +
				":\t\tvalue = 16'b" + instr.getBinaryCode() + ";");
		}
		writer.println("\t\tdefault:\t\tvalue = 16'b0000000000000000;");
		writer.close();
		System.out.println("Generated machine code to " + outputFileName);
	}

	public AvrParser() {}

	public AvrParser(String filename) {
		this.filename = filename;
	}
}
